//
//  AppDelegate.swift
//  Magnetometer
//
//  Created by Phidgets on 2018-02-13.
//  Copyright © 2018 Phidgets. All rights reserved.
//

import Cocoa
import Phidget22Swift

@NSApplicationMain
class AppDelegate: NSObject, NSApplicationDelegate {

    @IBOutlet weak var window: NSWindow!
    let ch = Magnetometer()
    @IBOutlet weak var phidgetInfoBox: PhidgetInfoBox!

    //GUI
    @IBOutlet weak var changeTriggerSlider: NSSlider!
    @IBOutlet weak var minChangeTriggerLabel: NSTextField!
    @IBOutlet weak var maxChangeTriggerLabel: NSTextField!
    @IBOutlet weak var changeTriggerLabel: NSTextField!
    @IBOutlet weak var dataIntervalSlider: NSSlider!
    @IBOutlet weak var minDataIntervalLabel: NSTextField!
    @IBOutlet weak var maxDataIntervalLabel: NSTextField!
    @IBOutlet weak var dataIntervalLabel: NSTextField!
    @IBOutlet weak var xLabel: NSTextField!
    @IBOutlet weak var yLabel: NSTextField!
    @IBOutlet weak var zLabel: NSTextField!
    @IBOutlet weak var timestampLabel: NSTextField!
    @IBOutlet weak var settingsBox: NSBox!
    @IBOutlet weak var dataBox: NSBox!
    @IBOutlet weak var calibrationStatusLabel: NSTextField!
    
    //Error Window
    @IBOutlet weak var errorEventWindow: NSPanel!
    @IBOutlet var errorEventLog: NSTextView!
    @IBOutlet weak var errorEventCount: NSTextField!
    var errorCounter = 0
    
    //Magnetometer Parameters
    @IBOutlet weak var parametersWindow: NSPanel!
    @IBOutlet weak var parametersText: NSTextField!
    
    
    
    func applicationDidFinishLaunching(_ aNotification: Notification) {
        do {
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
            
            let _ = ch.attach.addHandler(attach_handler)
            let _ = ch.detach.addHandler(detach_handler)
            let _ = ch.error.addHandler(error_handler)
            let _ = ch.magneticFieldChange.addHandler(fieldchange_handler)
            
            /*
             * Please review the Phidget22 channel matching documentation for details on the device
             * and class architecture of Phidget22, and how channels are matched to device features.
             */
            
            /*
             * Specifies the serial number of the device to attach to.
             * For VINT devices, this is the hub serial number.
             *
             * The default is any device.
             */
            // try ch.setDeviceSerialNumber(<YOUR DEVICE SERIAL NUMBER>)
            
            /*
             * For VINT devices, this specifies the port the VINT device must be plugged into.
             *
             * The default is any port.
             */
            // try ch.setHubPort(0)
            
            /*
             * Specifies which channel to attach to.  It is important that the channel of
             * the device is the same class as the channel that is being opened.
             *
             * The default is any channel.
             */
            // try ch.setChannel(0)
            
            /*
             * In order to attach to a network Phidget, the program must connect to a Phidget22 Network Server.
             * In a normal environment this can be done automatically by enabling server discovery, which
             * will cause the client to discovery and connect to available servers.
             *
             * To force the channel to only match a network Phidget, set remote to 1.
             */
            
            // try Net.enableServerDiscovery(serverType: .deviceRemote)
            // try ch.setIsRemote(true)
            
            /*
             openCmdLine allows command line arguments to be passed to this program in order to
             configure channel matching.
             If you have modified the code and added your own channel matching, comment out
             the openCmdLine call, and simply replace it with the following line:
             */
            //try ch.open()
            
            try phidgetInfoBox.openCmdLine(phid: ch as Phidget)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func applicationShouldTerminateAfterLastWindowClosed(_ sender: NSApplication) -> Bool {
        return true
    }
    
    func applicationWillTerminate(_ aNotification: Notification) {
        //Safely close example
        do{
            print("Closing")
            ch.attach.removeAllHandlers()
            ch.detach.removeAllHandlers()
            ch.error.removeAllHandlers()
            ch.magneticFieldChange.removeAllHandlers()
            try ch.close()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Event handlers
    func attach_handler(sender: Phidget){
        do{
            let attachedDevice = sender as! Magnetometer
            let minChangeTrigger = try attachedDevice.getMinMagneticFieldChangeTrigger()
            let maxChangeTrigger = try attachedDevice.getMaxMagneticFieldChangeTrigger()
            let changeTrigger = try attachedDevice.getMagneticFieldChangeTrigger()
            let minDataInterval = try attachedDevice.getMinDataInterval()
            let maxDataInterval = try attachedDevice.getMaxDataInterval()
            let dataInterval = try attachedDevice.getDataInterval()
            
            DispatchQueue.main.async {
                self.phidgetInfoBox.fillPhidgetInfo(device: sender)
                self.settingsBox.isHidden = false
                self.dataBox.isHidden = false
                self.calibrationStatusLabel.stringValue = ""
                
                self.minChangeTriggerLabel.stringValue = String(format:"%.3f", minChangeTrigger)
                self.maxChangeTriggerLabel.stringValue = String(format:"%.3f", maxChangeTrigger)
                self.changeTriggerLabel.stringValue = String(format:"%.3f", changeTrigger)
                self.changeTriggerSlider.doubleValue = changeTrigger
                
                self.changeTriggerSlider.minValue = minChangeTrigger
                self.changeTriggerSlider.maxValue = maxChangeTrigger
                self.changeTriggerSlider.doubleValue = changeTrigger
                
                self.minDataIntervalLabel.stringValue = String(format:"%d",minDataInterval)
                self.maxDataIntervalLabel.stringValue = String(format:"%d",maxDataInterval)
                self.dataIntervalLabel.stringValue = String(format:"%d",dataInterval)
                
                self.dataIntervalSlider.minValue = Double(minDataInterval)
                self.dataIntervalSlider.maxValue = Double(maxDataInterval)
                self.dataIntervalSlider.doubleValue = Double(dataInterval)
                
                //Adjusting window height
                var frame = self.window.frame
                frame.size.height = self.window.maxSize.height
                self.window.setFrame(frame, display: true, animate: false)
                self.window.center()
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func detach_handler(sender: Phidget){
        DispatchQueue.main.async {
            self.phidgetInfoBox.fillPhidgetInfo(device: nil)
            self.settingsBox.isHidden = true
            self.dataBox.isHidden = true
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
        }
    }
    
    func error_handler(sender: Phidget, data: (code: ErrorEventCode, description: String)){
        outputError(errorDescription: data.description, errorCode: data.code.rawValue)
    }
    
    func fieldchange_handler(sender: Magnetometer, data: (magneticField: [Double], timestamp: Double)){
        DispatchQueue.main.async {
            self.xLabel.stringValue = String(format:"%0.3f G", data.magneticField[0])
            self.yLabel.stringValue = String(format:"%0.3f G", data.magneticField[1])
            self.zLabel.stringValue = String(format:"%0.3f G", data.magneticField[2])
            self.timestampLabel.stringValue = String(format:"%.0f ms", data.timestamp)
        }
    }
    
    //MARK: GUI Controls
    @IBAction func setChangeTrigger(_ sender: Any) {
        let slider = sender as! NSSlider
        do{
            try ch.setMagneticFieldChangeTrigger(slider.doubleValue)
            changeTriggerLabel.stringValue = String(format:"%.3f G", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setDataInterval(_ sender: Any) {
        let slider = sender as! NSSlider
        do{
            try ch.setDataInterval(UInt32(slider.intValue))
            dataIntervalLabel.stringValue = String(format:"%d ms", UInt32(slider.intValue))
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func openParametersWindow(_ sender: Any) {
        if(!parametersWindow.isVisible){
            parametersWindow.setIsVisible(true)
            
        }
    }
    
    @IBAction func setParameters(_ sender: Any) {
        do{
            let strArray = parametersText.stringValue.components(separatedBy: ",")
            if(strArray.count == 1){//empty
                try ch.resetCorrectionParameters()
                try ch.saveCorrectionParameters()
                calibrationStatusLabel.stringValue = "Un-Calibrated"
                calibrationStatusLabel.textColor = NSColor.red
            }
            else if(strArray.count != 13){
                errorAlert(errorStr: "Number of parameters must be 13.")
                calibrationStatusLabel.stringValue = "Calibration Error"
                calibrationStatusLabel.textColor = NSColor.red
            }
            else{
                var doubleArr = [Double]()
                for str in strArray{
                    doubleArr.append(Double(str)!)
                }
                try ch.setCorrectionParameters(magneticField: doubleArr[0], offset0: doubleArr[1], offset1: doubleArr[2], offset2: doubleArr[3], gain0: doubleArr[4], gain1: doubleArr[5], gain2: doubleArr[6], T0: doubleArr[7], T1: doubleArr[8], T2: doubleArr[9], T3: doubleArr[10], T4: doubleArr[11], T5: doubleArr[12])
                calibrationStatusLabel.stringValue = "Calibration Complete"
                calibrationStatusLabel.textColor = NSColor.green
            }
            parametersWindow.setIsVisible(false)
        } catch let err as PhidgetError {
            errorAlert(errorStr: err.description)
            calibrationStatusLabel.stringValue = "Calibration Error"
            calibrationStatusLabel.textColor = NSColor.red
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func errorAlert(errorStr: String){
        let alert = NSAlert()
        alert.messageText = errorStr
        alert.alertStyle = .warning
        alert.addButton(withTitle: "OK")
        alert.runModal()
    }
    
    @IBAction func cancelParameters(_ sender: Any) {
        parametersWindow.setIsVisible(false)
    }
    
    //MARK: Error
    func outputError(errorDescription:String, errorCode:UInt32){
        DispatchQueue.main.async {
            self.errorCounter += 1
            let outputString = NSAttributedString(string: String(format:"Error %d: %@\n", errorCode, errorDescription))
            self.errorEventLog.textStorage?.beginEditing()
            self.errorEventLog.textStorage?.append(outputString)
            self.errorEventLog.textStorage?.endEditing()
            self.errorEventCount.stringValue = "\(self.errorCounter)"
            if(!self.errorEventWindow.isVisible){
                self.errorEventWindow.setIsVisible(true)
                self.errorEventWindow.center()
            }
        }
        
    }
    
    @IBAction func clearError(_ sender: Any) {
        errorEventLog.textStorage?.setAttributedString(NSAttributedString(string: ""))
        errorCounter = 0
        errorEventCount.stringValue = "0"
    }
}

